/*
======================================================================
LOCK-MANAGER:	Version: 1.1 (94/01/19)

Copyright 1994:	Martin Weitzel,
		Darmstadt, Germany
		(martin@rent-a-guru.DE)
======================================================================
*/

/*
 * ISO-C standard header files
*/
#include <assert.h>
#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <varargs.h>
#include <unistd.h>

/*
 * UNIX system header files
*/
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/fcntl.h>
#include <sys/wait.h>

/*
 * External definitions (not in a header file)
*/
extern int open();
extern unsigned int alarm();
extern int setuid();

/*
 * Support for systems without `mkfifo'
 */
#ifdef HAS_NO_MKFIFO
#define mkfifo(p, m) (mknod((p), S_IFIFO|(m)))
#endif

/*
 * Program link names to control functionality (in case
 * of changes, Makefile must be changed accordingly)
*/
char LCKMAN[] = "lckman";	/* start demon and/or add facility */
char LCKREQ[] = "lckreq";	/* request a lock */
char LCKSHUT[] = "lckshut";	/* shut down demon/remove facility */
char LCKSTAT[] = "lckstat";	/* show status of facilities */

/*
 * Configurable constants
*/
char FIFO_DIR[] = "./lckm_tmp/";	/* where all the fifos go */
#define MAX_NAMELEN	12	/* max. length for facility names */
#define C_TIMEOUT	3	/* timeout for opening C-fifo */
#define R_TIMEOUT	20	/* timeout for opening R-fifo */

/*
 * Useful definitions
 */
#define NEW(t) ((t*)malloc(sizeof(t)))

/*
 * Shutup complaints by lint
*/
#ifdef lint
static va_list _va_dummy;
#define malloc(n) ((void *)0)
#undef va_start
#define va_start(x) (x=_va_dummy)
#undef va_arg
#define va_arg(x, t) ((t)0)
#undef va_end
#define va_end(x) (_va_dummy=x)
#endif

/*
 * Definitions for command data structures
*/
typedef struct command {	/* command send to lock manager */
	int type;			/* type of command */
#		define CMD_SERV '='		/* start service */
#		define CMD_REQ  '+'		/* request lock */
#		define CMD_REL  '-'		/* release lock */
#		define CMD_STAT '?'		/* show status */
#		define CMD_SHUT '.'		/* shut down */
#		define CMD_VERB '$'		/* toggle verbosity */
	char lockname[MAX_NAMELEN+1];		/* name of lock */
	int rpid;			/* PID of requestor */
} COMMAND;

/*
 * Global variables
*/
char *prog;			/* name of the program */
int verbose;			/* verbosity flag */		
char *c_fifo;			/* name of control fifo */
int cfd = -1;			/* FD of control fifo */

/*!
----------------------------------------------------------------------
	Following are some miscalaneous functions called as
	necessary from various parts of the program.
----------------------------------------------------------------------
*/

/*
 * Print message and optionally terminate program
*/
/*VARARGS*/
void failed(va_alist)
	va_dcl			/* fixed args "rc" and "fmt" (see below) */
{
	int rc;				/* return status of program */
	char *fmt;			/* printf format for message */
	va_list ap;			/* arg list iterator */

	/* format error message */
	va_start(ap);
	rc = va_arg(ap, int);
	fmt = va_arg(ap, char *);
	(void) fprintf(stderr, "%s: %s ", prog,
		rc ? "[FATAL]" : "[WARNING]");
	(void) vfprintf(stderr, fmt, ap);
	(void) fprintf(stderr, "\n");
	(void) fflush(stderr);
	va_end(ap);

	/* terminate program if "rc" is not 0 */
	if (rc > 0) exit(rc);
	if (rc < 0) abort();

	/* continue if "rc" is 0 */
}

/*
 * Standard messages
*/
static char _F_MALLOC[] = "out of malloc-space in %s";
#define FAILED_MALLOC(w) failed(-1, _F_MALLOC, w)

/*
 * Print a message only in verbose mode
*/
/*VARARGS*/
void inform(va_alist)
	va_dcl			/* fixed arg is "fmt" (see below) */
{


	char *fmt;			/* printf format for message */
	va_list ap;			/* arg list iterator */

	if (verbose) {
		/* format message */
		va_start(ap);
		fmt = va_arg(ap, char *);
		(void) vprintf(fmt, ap);
		va_end(ap);
		(void) fflush(stdout);
	}
}

/*
 * Rudimentary string processing support:
*/
char *mknstr(num)		/* -- static string from */
	int num;		/* number */
{
	static char *cp;		/* pointer to return space */
	size_t len = 0;			/* length of return space */
	int v = num;			/* tmp to determine length */

	assert(num >= 0);

	/* determine length */
	do ++len; while (v /= 10);

	/* free old and allocate new space */
	if (cp != (char *)0)
		free(cp);
	if ((cp = malloc(len+1)) == (char *)0)
		FAILED_MALLOC("mknstr");

	/* convert number to string */
	cp[len] = '\0';
	do {
		assert(len != 0);
		cp[--len] = (num % 10) + '0';
	} while (num /= 10);

	return cp;
}

/*VARARGS*/
char *pstrcat(va_alist)		/* -- dynamic string from parts */
	va_dcl			/* char pointer args, (char *)0-terminated */
{
	char *cp;			/* tmp to access args */
	char *res;			/* tmp for result */
	size_t len = 1;			/* length of all args (plus '\0') */
	va_list ap;			/* arg list iterator */

	/* determine total length */
	va_start(ap);
	while ((cp = va_arg(ap, char *)) != (char *)0)
		len += strlen(cp);
	va_end(ap);

	/* allocate space and initialize */
	if ((res = malloc(len)) == (char *)0)
		FAILED_MALLOC("pstrcat");
	res[0] = '\0';

	/* concatenate all arg strings */
	va_start(ap);
	while ((cp = va_arg(ap, char *)) != (char *)0)
		(void) strcat(res, cp);
	va_end(ap);

	return res;
}

void pstrfree(strp)		/* -- free dynamic string */
	char **strp;		/* address(!) of pointer to string */
{
	if (*strp)
		free(*strp);
	*strp = (char *)0;
}

/*
 * Build name of R-fifo from PID of requestor
*/
char *r_fifoname(p)
	int p;			/* PID */
{
	return pstrcat(FIFO_DIR, "R-", mknstr(p), (char *)0);
}

/*
 * Build name of C-fifo from name of facility
*/
char *c_fifoname(lockname)
	char *lockname;		/* name of facitily */
{
	return pstrcat(FIFO_DIR, "C-", lockname, (char *)0);
}

/*
 * Write command to C-fifo:
*/
void sendpcmd(cmdp)		/* -- with parameters */
	COMMAND *cmdp;		/* initialized command structure */
{
	assert(cfd != -1);

	if (write(cfd, cmdp, sizeof *cmdp) != sizeof *cmdp)
		failed(-1, "error %d writing to control fifo", errno);
}
void sendscmd(type)		/* -- without parameters */
	int type;		/* command type */
{
	static COMMAND cmd;		/* command structure */
	cmd.type = type;
	sendpcmd(&cmd);
}

/*
 * Check and conditionally truncate name of facility
*/
void ckname(lockname)
	char *lockname;		/* name of facility */
{
	if (strlen(lockname) > (unsigned)MAX_NAMELEN) {
		lockname[MAX_NAMELEN] = '\0';
		failed(0, "lock name trunctated to %s", lockname);
	}
}

/*
 * Open file with time-out
*/
/*ARGSUSED*/
void nothing(sig)
	int sig;
{
	/*empty*/
}
int tmopen(name, mode, tout)	/* -- similar to open(2) */
	char *name;		/* file name */
	int mode;		/* open mode */
	unsigned tout;		/* time-out */
{
	int fd;				/* tmp for resulting FD */
	void (*oldsig)();		/* tmp for old sig handler */

	oldsig = signal(SIGALRM, nothing);
	(void) alarm(tout);
	fd = open(name, mode);
	(void) alarm(0);
	(void) signal(SIGALRM, oldsig);

	return fd;
}

/*
 * Request some operation from the demon
*/
void requestOP(type, name)
	int type;		/* type of operation */
	char *name;		/* (opt.) facility name */
{
	assert(c_fifo != (char *)0);

	if (name == (char *)0 || name[0] == '\0')
		sendscmd(type);
	else {
		COMMAND cmd;
		ckname(name);
		cmd.type = type;
		(void) strcpy(cmd.lockname, name);
		sendpcmd(&cmd);
	}
}

/*!
----------------------------------------------------------------------
	The following functions implement the lock manager's
	kernal data structures and algorithms.  This part
	typically runs in the background waiting for commands
	to read from the C-fifo.
----------------------------------------------------------------------
*/

typedef struct request {	/* chain of requests */
	int rpid;			/* PID of requestor */
	struct request *next;		/* next request in chain */
} REQUEST;

typedef struct lock {		/* chain of facilities */
	char lockname[MAX_NAMELEN+1];	/* name of this facility */
	int serviced;			/* service indicator */
	int gpid;			/* PID of granter */
	REQUEST *req;			/* chain of requests */
	struct lock *next;		/* next lock in chain */
} LOCK;

LOCK *locklist;			/* anchor of list */

/*
 * Append request to end of queue
*/
void addrequest(rpid, rpp)
	int rpid;		/* PID of requestor */
	REQUEST **rpp;		/* pointer to anchor of request queue */
{
	/* find end of chain */
	while (*rpp != (REQUEST *)0)
		rpp = &(*rpp)->next;

	/* allocate space */
	*rpp = NEW(REQUEST);
	if (*rpp == (REQUEST *)0)
		FAILED_MALLOC("addrequest");

	/* initialize data */
	(*rpp)->rpid = rpid;
	(*rpp)->next = (REQUEST *)0;
}

/*
 * Delete request from begin of queue
*/
void delrequest(rpp)
	REQUEST **rpp;		/* pointer to anchor of request queue */
{
	REQUEST *rp = *rpp;	/* pointer to entry to free */

	assert(*rpp != (REQUEST *)0);

	/* remove entry and free space */
	*rpp = rp->next;
	free(rp);
}

/*
 * Make entry into the list of facilities
*/
void mklock(lockname)
	char *lockname;		/* name of facility */
{
	/* allocate space */
	LOCK *lp = NEW(LOCK);
	if (lp == (LOCK *)0)
		FAILED_MALLOC("mklock");

	/* initialize data */
	(void) strcpy(lp->lockname, lockname);
	lp->serviced = 1;
	lp->gpid = -1;
	lp->req = (REQUEST *)0;

	/* prepend to existing list */
	lp->next = locklist;
	locklist = lp;
}

/*
 * Locate entry in the list of facilities:
*/
LOCK *gtlockn(lockname)		/* -- by */
	char *lockname;		/* name of facility */
{
	LOCK *lp = locklist;
	while (lp != (LOCK *)0) {
		if (!strcmp(lockname, lp->lockname))
			break;
		lp = lp->next;
	}
	return lp;
}
LOCK *gtlockp(gpid)		/* -- by */
	int gpid;		/* PID of grantor */
{
	LOCK *lp = locklist;
	while (lp != (struct  lock *)0) {
		if (gpid == lp->gpid)
			break;
		lp = lp->next;
	}
	return lp;
}

/*
 * Remove entry from the list of facilities
*/
void rmlock(lp)
	LOCK *lp;		/* pointer to entry to remove */
{
	LOCK **lpp = &locklist;		/* pointer(!) to entry pointer */

	assert (lp != (LOCK *)0);
	assert (lp->serviced == 0);
	assert (lp->gpid == -1);
	assert (lp->req == (REQUEST *)0);

	/* traverse list to find given entry */
	while (*lpp != lp) {
		assert (*lpp != (LOCK *)0);
		lpp = &((*lpp)->next);
	}

	/* remove entry and free space */
	*lpp = lp->next;
	free(lp);
}

/*
 * Grant a lock
*/
void grantlock(rpid, gpidp)
	int rpid;		/* PID of requestor */
	int *gpidp;		/* pointer to PID of grantor */
{
	assert(rpid != -1);

	/* fork grantor */
	if ((*gpidp = fork()) == 0) {

		/* connect to R-fifo */
		char *r_fifo = r_fifoname(rpid);
		int rfd = tmopen(r_fifo, O_RDONLY, R_TIMEOUT);
		(void) unlink(r_fifo);
		pstrfree(&r_fifo);

		/* wait for requestor to vanish */
		(void) signal(SIGTERM, nothing);
		if (rfd != -1) {
			char dummy;
			while (read(rfd, &dummy, 1) > 0)
				; /*empty*/
			(void) close(rfd);
		}

		/* message release and terminate */
		sendscmd(CMD_REL);
		exit(0);
		/*NOTREACHED*/
	}
	if (*gpidp == -1)
		failed(0, "error %d on fork", errno);
}

/*
 * Stop servicing a facility
*/
void unservlock(lp)
	LOCK *lp;		/* pointer to facility */
{
	assert(lp != (LOCK *)0);

	inform("stop servicing <%s> -- ", lp->lockname);
	if (lp->serviced) {
		char *cs_fifo = c_fifoname(lp->lockname);
		(void) unlink(cs_fifo);
		pstrfree(&cs_fifo);
		lp->serviced = 0;
		if (lp->gpid != -1)
			inform("OK (initiated)\n");
		else {
			rmlock(lp);
			inform("OK (shut down)\n");
		}
	}
	else 
		inform("FAILED (servicing already stopped)");
}

/*
 * Handle start or re-start of facility
*/
void startService(name)
	char *name;		/* name of facility */
{
	LOCK *lp = gtlockn(name);	/* pointer to facility */
	char *f_fifo = c_fifoname(name);/* name of facility-dependend fifo */

	assert(c_fifo != (char *)0);

	inform("start servicing <%s> -- ", name);
	if (lp != (LOCK *)0)
		if (lp->serviced)
			inform("FAILED (already running)\n");
		else {
			lp->serviced = 1;
			if (link(c_fifo, f_fifo) != -1)
				inform("OK (shut down canceled)\n");
			else
				inform("FAILED (cannot re-link)\n");
		}
	else {
		/* link generic C-fifo to facility-dependend name */
		if (unlink(f_fifo) != -1)
			failed(0, "dead(?) control fifo %s removed", f_fifo);
		if (link(c_fifo, f_fifo) != -1)
			mklock(name);
		else
			failed(0, "error %d linking %s to %s", errno,
				c_fifo, f_fifo);
		inform("OK\n");
	}
	pstrfree(&f_fifo);
}

/*
 * Handle request for a lock
*/
void doRequest(lockname, rpid)
	char *lockname;		/* name of facility */
	int rpid;		/* PID of requestor */
{
	LOCK *lp = gtlockn(lockname);	/* pointer to facility */

	inform("request for <%s> by R-%d -- ", lockname, rpid);
	if (lp == (LOCK *)0)
		inform("FAILED (no such facility)\n");
	else if (!lp->serviced)
		inform("FAILED (not serviced any more)");
	else if (lp->gpid == -1) {
		grantlock(rpid, &lp->gpid);
		inform("OK (G-%d)\n", lp->gpid);
	}
	else {
		addrequest(rpid, &lp->req);
		inform("OK (queued)\n");
	}
}

/*
 * Handle release of a lock
*/
void doRelease()
{
	int gpid;			/* PID of grantor */
	LOCK *lp;			/* pointer to released facility */

	while ((gpid = wait((int *)0)) == -1) {
		if (errno != EINTR)
			failed(-1, "error %d waiting for child", errno);
	}

	/* find release facility */
	lp = gtlockp(gpid);
	assert(lp != (LOCK *)0);
	inform("release of <%s> by G-%d -- ",
		lp->lockname, gpid);

	if (lp->req != (REQUEST *)0) {
		/* grant to next requestor */
		grantlock(lp->req->rpid, &lp->gpid);
		inform("OK (G-%d)\n", lp->gpid);
		delrequest(&lp->req);
	}
	else {
		/* mark facility as available */
		lp->gpid = -1;

		if (lp->serviced)
			inform("OK (available)\n");
		else {
			rmlock(lp);
			inform("OK (finally shut down)\n");
		}
	}
}

/*
 * Handle status display of all facilities
*/
void showLocks()
{
	LOCK *lp = locklist;		/* iterator for facility list */
	while (lp != (LOCK *)0) {
		REQUEST *rp = lp->req;	/* iterator for request list */
		(void) printf("facility <%s> %s", lp->lockname,
			lp->serviced ? "serviced" : "shuting down");
		if (lp->gpid != -1)
			(void) printf(" (G-%d)", lp->gpid);
		while (rp != (REQUEST *)0) {
			(void) printf(" R-%d", rp->rpid);
			rp = rp->next;
		}
		(void) printf("\n");
		lp = lp->next;
	}
	(void) fflush(stdout);
}

/*
 * Handle shut down of facility
*/
void shutService(name)
	char *name;		/* (begin of) name of facility */
{
	LOCK *lp = locklist;
	while (lp != (LOCK *)0) {
		LOCK *lpn = lp->next;
		if (!strncmp(lp->lockname, name, strlen(name)))
			unservlock(lp);
		lp = lpn;
	}
}

int run;			/* global run flag */

/*
 * Manage request for and release of locks
 * and administrative commands
*/
void stop(sig)			/* -- signal handler */
	int sig;		/* signal number */
{
	requestOP(CMD_SHUT, (char *)0);
	run = -sig;
}
void LockManager()		/* -- command dispatch demon */
{
	/* create control fifo */
	(void) umask(0);
	if (mkfifo(c_fifo, 0620) == -1)
		failed(-1, "error %d creating %s", errno, c_fifo);

	/* put demon in the background */
	switch (fork()) {
	case -1:
		failed(201, "error %d forking demon process", errno);
		/*NOTREACHED*/
	case 0:
		/* open control fifo in demon */
		if ((cfd = open(c_fifo, O_RDWR)) == -1)
			failed(-1, "error %d opening %s", errno, c_fifo);
		break;
	default:
		/* foreground task continues here */
		return;
	}

	/* prepare continuoues operation */
	run = 1;
	(void) signal(SIGINT, SIG_IGN);
	(void) signal(SIGQUIT, SIG_IGN);
	(void) signal(SIGHUP, stop);
	(void) signal(SIGTERM, stop);
	
	while (run > 0 || locklist != (LOCK *)0) {
		COMMAND cmd;

		/* wait for command to read from control fifo */
		if (read(cfd, &cmd, sizeof cmd) != sizeof cmd) {
			if (errno == EINTR && run < 0) {
				cmd.type = CMD_SHUT;
				cmd.lockname[0] = '\0';
			}
			else failed(-1, "error %d reading %s", errno,
				c_fifo);
		}

		/* dispatch command to handler */
		switch (cmd.type) {
		case CMD_SERV:
			startService(cmd.lockname);
			break;
		case CMD_REQ:
			doRequest(cmd.lockname, cmd.rpid);
			break;
		case CMD_REL:
			doRelease();
			break;
		case CMD_STAT:
			showLocks();
			break;
		case CMD_VERB:
			verbose ^= 1;
			break;
		case CMD_SHUT:
			shutService(cmd.lockname);
			if (cmd.lockname[0] == '\0' && run > 0)
				run = 0;
			break;
		default:
			failed(0, "bad command read from %s", c_fifo);
		}
	}

	/* remove control fifo */
	(void) unlink(c_fifo);
	exit(run ? 128 - run : 0);
}

/*!
----------------------------------------------------------------------
	The following part implement the various commands that
	typically run in the foreground and connect to the demon
	via the control fifo.
----------------------------------------------------------------------
*/

/*
 * Request a lock from the manager
*/
void requestLock(lockname)
	char *lockname;			/* name of facility */
{
	static COMMAND req = {CMD_REQ};	/* command structure to send */
	char *r_fifo;			/* name of R-fifo */
	ckname(lockname);

	/* check if facility is serviced */
	c_fifo = c_fifoname(lockname);
	if ((cfd = open(c_fifo, O_WRONLY)) == -1)
		failed(3, "lock %s not serviced", lockname);

	/* complete command structure */
	req.rpid = getppid();
	(void) strcpy(req.lockname, lockname);
	r_fifo = r_fifoname(req.rpid);

	/* create R-fifo */
	(void) umask(0);
	if (mkfifo(r_fifo, 0660) == -1)
		failed(-1, "cannot create R-fifo %s", r_fifo);

	/* send command to manager */
	sendpcmd(&req);

	/* print name of R-fifo */
	(void) printf("%s", r_fifo);

	/* miscalaneous cleanup */
	pstrfree(&r_fifo);
	pstrfree(&c_fifo);
}

/*
 * Establish control fifo, optionally start demon
*/
void setupChannel(runmgr)
	int runmgr;		/* indicator to start demon */
{
	/* try to connect to control fifo */
	c_fifo = c_fifoname("");
	while ((cfd = tmopen(c_fifo, O_WRONLY, C_TIMEOUT)) == -1) {
		if (runmgr) {
			if (errno == EINTR) {
				/* demon not responding in time */
				failed(0, "demon died, trying restart");
				(void) unlink(c_fifo);
				errno = ENOENT;
			}
			if (errno == ENOENT) {
				/* no control fifo, demon not running */
				runmgr = 0;
				LockManager();
				continue;
			}
		}
		if (errno == EINTR || errno == ENOENT)
			failed(4, "lock manager not running");
		failed(-1, "error %d opening %s", errno, c_fifo);
	}
}

/*
 * Execute functionality depending on program name
*/
int progis(cmpto)		/* -- compare basename of program */
	char *cmpto;		/* name to compare to */
{
	char *cp = prog + strlen(prog);
	while (cp > prog && cp[-1] != '/')
		--cp;
	return !strcmp(cp, cmpto);
}
int main(argc, argv)		/* -- main program */
	int argc;		/* number of command line args */
	char **argv;		/* array of command line args */
{
	/* store program name */
	prog = *argv;

	if (progis(LCKREQ)) {
		if (argc != 2) {
			(void) fprintf(stderr, "Usage: %s facility\n", prog);
			return 2;
		}
		if (setuid(getuid()) == -1)
			failed(-1, "error %d on setuid(%d)", errno,
				getuid());
		requestLock(*++argv);
		return 0;
	}

	/*
	 * check for priviledged user
	*/
	if (getuid() == 0) {
		 if (setgid(getegid()) == -1)
			failed(-1, "error %d changing gids", errno);
	}
	if (getgid() != getegid())
		failed(9, "not authorized for this function");

	/*
	 * access to control fifo / start demon
	*/
	setupChannel(progis(LCKMAN));

	if (progis(LCKMAN)) {
		if (argc > 1 && !strcmp(argv[1], "-v")) {
			requestOP(CMD_VERB, (char *)0);
			++argv;
		}
		while (*++argv)
			requestOP(CMD_SERV, *argv);
		return 0;
	}

	if (progis(LCKSTAT)) {
		if (argc != 1) {
			(void) fprintf(stderr, "Usage: %s (no args)\n", prog);
			return 2;
		}
		requestOP(CMD_STAT, (char *)0);
		return 0;
	}

	if (progis(LCKSHUT)) {
		if (argc == 1)
			requestOP(CMD_SHUT, (char *)0);
		else while (*++argv)
			requestOP(CMD_SHUT, *argv);
		return 0;
	}

	failed(1, "no functionality defined");
	/*NOTREACHED*/
}
